@@ -4,6 +4,9 @@
 
 """
 
+# remove when we don't support py38 anymore
+from __future__ import annotations
+from typing import Any
 import argparse
 import os
 import shutil
@@ -11,7 +14,13 @@ import logging
 import configparser
 import sys
 
-from jinja2 import Environment, ChoiceLoader, FileSystemLoader, PackageLoader
+from jinja2 import (
+    Environment,
+    ChoiceLoader,
+    FileSystemLoader,
+    PackageLoader,
+    Template,
+)
 import feedgenerator
 
 from blag.markdown import markdown_factory, convert_markdown
@@ -21,12 +30,12 @@ from blag.quickstart import quickstart
 
 logger = logging.getLogger(__name__)
 logging.basicConfig(
-        level=logging.INFO,
-        format='%(asctime)s %(levelname)s %(name)s %(message)s',
+    level=logging.INFO,
+    format='%(asctime)s %(levelname)s %(name)s %(message)s',
 )
 
 
-def main(args=None):
+def main(arguments: list[str] | None = None) -> None:
     """Main entrypoint for the CLI.
 
     This method parses the CLI arguments and executes the respective
@@ -34,20 +43,24 @@ def main(args=None):
 
     Parameters
     ----------
-    args : list[str]
+    arguments
         optional parameters, used for testing
 
     """
-    args = parse_args(args)
+    args = parse_args(arguments)
+    # set loglevel
+    if args.verbose:
+        logger.setLevel(logging.DEBUG)
+    logger.debug(f"This is blag {__VERSION__}.")
     args.func(args)
 
 
-def parse_args(args=None):
+def parse_args(args: list[str] | None = None) -> argparse.Namespace:
     """Parse command line arguments.
 
     Parameters
     ----------
-    args : List[str]
+    args
         optional parameters, used for testing
 
     Returns
@@ -59,85 +72,99 @@ def parse_args(args=None):
     parser.add_argument(
         '--version',
         action='version',
-        version='%(prog)s '+__VERSION__,
+        version='%(prog)s ' + __VERSION__,
+    )
+    parser.add_argument(
+        '-v',
+        '--verbose',
+        action='store_true',
+        help='Verbose output.',
     )
 
     commands = parser.add_subparsers(dest='command')
     commands.required = True
 
     build_parser = commands.add_parser(
-            'build',
-            help='Build website.',
+        'build',
+        help='Build website.',
     )
     build_parser.set_defaults(func=build)
     build_parser.add_argument(
-            '-i', '--input-dir',
-            default='content',
-            help='Input directory (default: content)',
+        '-i',
+        '--input-dir',
+        default='content',
+        help='Input directory (default: content)',
     )
     build_parser.add_argument(
-            '-o', '--output-dir',
-            default='build',
-            help='Ouptut directory (default: build)',
+        '-o',
+        '--output-dir',
+        default='build',
+        help='Ouptut directory (default: build)',
     )
     build_parser.add_argument(
-            '-t', '--template-dir',
-            default='templates',
-            help='Template directory (default: templates)',
+        '-t',
+        '--template-dir',
+        default='templates',
+        help='Template directory (default: templates)',
     )
     build_parser.add_argument(
-            '-s', '--static-dir',
-            default='static',
-            help='Static directory (default: static)',
+        '-s',
+        '--static-dir',
+        default='static',
+        help='Static directory (default: static)',
     )
 
     quickstart_parser = commands.add_parser(
-            'quickstart',
-            help="Quickstart blag, creating necessary configuration.",
+        'quickstart',
+        help="Quickstart blag, creating necessary configuration.",
     )
     quickstart_parser.set_defaults(func=quickstart)
 
     serve_parser = commands.add_parser(
-            'serve',
-            help="Start development server.",
+        'serve',
+        help="Start development server.",
     )
     serve_parser.set_defaults(func=serve)
     serve_parser.add_argument(
-            '-i', '--input-dir',
-            default='content',
-            help='Input directory (default: content)',
+        '-i',
+        '--input-dir',
+        default='content',
+        help='Input directory (default: content)',
     )
     serve_parser.add_argument(
-            '-o', '--output-dir',
-            default='build',
-            help='Ouptut directory (default: build)',
+        '-o',
+        '--output-dir',
+        default='build',
+        help='Ouptut directory (default: build)',
     )
     serve_parser.add_argument(
-            '-t', '--template-dir',
-            default='templates',
-            help='Template directory (default: templates)',
+        '-t',
+        '--template-dir',
+        default='templates',
+        help='Template directory (default: templates)',
     )
     serve_parser.add_argument(
-            '-s', '--static-dir',
-            default='static',
-            help='Static directory (default: static)',
+        '-s',
+        '--static-dir',
+        default='static',
+        help='Static directory (default: static)',
     )
 
     return parser.parse_args(args)
 
 
-def get_config(configfile):
+def get_config(configfile: str) -> configparser.SectionProxy:
     """Load site configuration from configfile.
 
     Parameters
     ----------
-    configfile : str
+    configfile
         path to configuration file
 
 
     Returns
     -------
-    dict
+    configparser.SectionProxy
 
     """
     config = configparser.ConfigParser()
@@ -157,7 +184,10 @@ def get_config(configfile):
     return config['main']
 
 
-def environment_factory(template_dir=None, globals_=None):
+def environment_factory(
+    template_dir: str | None = None,
+    globals_: dict[str, object] | None = None,
+) -> Environment:
     """Environment factory.
 
     Creates a Jinja2 Environment with the default templates and
@@ -167,8 +197,9 @@ def environment_factory(template_dir=None, globals_=None):
 
     Parameters
     ----------
-    template_dir : str
-    globals_ : dict
+    template_dir
+        directory containing the templates
+    globals_
 
     Returns
     -------
@@ -177,7 +208,7 @@ def environment_factory(template_dir=None, globals_=None):
     """
     # first we try the custom templates, and fall back the ones provided
     # by blag
-    loaders = []
+    loaders: list[FileSystemLoader | PackageLoader] = []
     if template_dir:
         loaders.append(FileSystemLoader([template_dir]))
     loaders.append(PackageLoader('blag', 'templates'))
@@ -187,7 +218,7 @@ def environment_factory(template_dir=None, globals_=None):
     return env
 
 
-def build(args):
+def build(args: argparse.Namespace) -> None:
     """Build the site.
 
     This is blag's main method that builds the site, generates the feed
@@ -195,15 +226,16 @@ def build(args):
 
     Parameters
     ----------
-    args : argparse.Namespace
+    args
 
     """
     os.makedirs(f'{args.output_dir}', exist_ok=True)
     convertibles = []
     for root, dirnames, filenames in os.walk(args.input_dir):
         for filename in filenames:
-            rel_src = os.path.relpath(f'{root}/{filename}',
-                                      start=args.input_dir)
+            rel_src = os.path.relpath(
+                f'{root}/{filename}', start=args.input_dir
+            )
             # all non-markdown files are just copied over, the markdown
             # files are converted to html
             if rel_src.endswith('.md'):
@@ -211,14 +243,17 @@ def build(args):
                 rel_dst = rel_dst[:-3] + '.html'
                 convertibles.append((rel_src, rel_dst))
             else:
-                shutil.copy(f'{args.input_dir}/{rel_src}',
-                            f'{args.output_dir}/{rel_src}')
+                shutil.copy(
+                    f'{args.input_dir}/{rel_src}',
+                    f'{args.output_dir}/{rel_src}',
+                )
         for dirname in dirnames:
             # all directories are copied into the output directory
             path = os.path.relpath(f'{root}/{dirname}', start=args.input_dir)
             os.makedirs(f'{args.output_dir}/{path}', exist_ok=True)
 
     # copy static files over
+    logger.info('Copying static files.')
     if os.path.exists(args.static_dir):
         shutil.copytree(args.static_dir, args.output_dir, dirs_exist_ok=True)
 
@@ -241,7 +276,8 @@ def build(args):
     )
 
     generate_feed(
-        articles, args.output_dir,
+        articles,
+        args.output_dir,
         base_url=config['base_url'],
         blog_title=config['title'],
         blog_description=config['description'],
@@ -251,8 +287,13 @@ def build(args):
     generate_tags(articles, tags_template, tag_template, args.output_dir)
 
 
-def process_markdown(convertibles, input_dir, output_dir,
-                     page_template, article_template):
+def process_markdown(
+    convertibles: list[tuple[str, str]],
+    input_dir: str,
+    output_dir: str,
+    page_template: Template,
+    article_template: Template,
+) -> tuple[list[tuple[str, dict[str, Any]]], list[tuple[str, dict[str, Any]]]]:
     """Process markdown files.
 
     This method processes the convertibles, converts them to html and
@@ -263,16 +304,17 @@ def process_markdown(convertibles, input_dir, output_dir,
 
     Parameters
     ----------
-    convertibles : List[Tuple[str, str]]
+    convertibles
         relative paths to markdown- (src) html- (dest) files
-    input_dir : str
-    output_dir : str
-    page_template, archive_template : jinja2 template
+    input_dir
+    output_dir
+    page_template, archive_template
         templats for pages and articles
 
     Returns
     -------
-    articles, pages : List[Tuple[str, Dict]]
+    list[tuple[str, dict[str, Any]]], list[tuple[str, dict[str, Any]]]
+        articles and pages
 
     """
     logger.info("Converting Markdown files...")
@@ -281,7 +323,8 @@ def process_markdown(convertibles, input_dir, output_dir,
     articles = []
     pages = []
     for src, dst in convertibles:
-        logger.info(f'Processing {src}')
+        logger.debug(f'Processing {src}')
+
         with open(f'{input_dir}/{src}', 'r') as fh:
             body = fh.read()
 
@@ -307,37 +350,37 @@ def process_markdown(convertibles, input_dir, output_dir,
 
 
 def generate_feed(
-        articles,
-        output_dir,
-        base_url,
-        blog_title,
-        blog_description,
-        blog_author,
-):
+    articles: list[tuple[str, dict[str, Any]]],
+    output_dir: str,
+    base_url: str,
+    blog_title: str,
+    blog_description: str,
+    blog_author: str,
+) -> None:
     """Generate Atom feed.
 
     Parameters
     ----------
-    articles : list[list[str, dict]]
+    articles
         list of relative output path and article dictionary
-    output_dir : str
+    output_dir
         where the feed is stored
-    base_url : str
+    base_url
         base url
-    blog_title : str
+    blog_title
         blog title
-    blog_description : str
+    blog_description
         blog description
-    blog_author : str
+    blog_author
         blog author
 
     """
     logger.info('Generating Atom feed.')
     feed = feedgenerator.Atom1Feed(
-            link=base_url,
-            title=blog_title,
-            description=blog_description,
-            feed_url=base_url + 'atom.xml',
+        link=base_url,
+        title=blog_title,
+        description=blog_description,
+        feed_url=base_url + 'atom.xml',
     )
 
     for dst, context in articles:
@@ -358,16 +401,20 @@ def generate_feed(
         feed.write(fh, encoding='utf8')
 
 
-def generate_archive(articles, template, output_dir):
+def generate_archive(
+    articles: list[tuple[str, dict[str, Any]]],
+    template: Template,
+    output_dir: str,
+) -> None:
     """Generate the archive page.
 
     Parameters
     ----------
-    articles : list[list[str, dict]]
+    articles
         List of articles. Each article has the destination path and a
         dictionary with the content.
-    template : jinja2.Template instance
-    output_dir : str
+    template
+    output_dir
 
     """
     archive = []
@@ -381,46 +428,52 @@ def generate_archive(articles, template, output_dir):
         fh.write(result)
 
 
-def generate_tags(articles, tags_template, tag_template, output_dir):
+def generate_tags(
+    articles: list[tuple[str, dict[str, Any]]],
+    tags_template: Template,
+    tag_template: Template,
+    output_dir: str,
+) -> None:
     """Generate the tags page.
 
     Parameters
     ----------
-    articles : list[list[str, dict]]
+    articles
         List of articles. Each article has the destination path and a
         dictionary with the content.
-    tags_template, tag_template : jinja2.Template instance
-    output_dir : str
+    tags_template, tag_template
+    output_dir
 
     """
     logger.info("Generating Tag-pages.")
     os.makedirs(f'{output_dir}/tags', exist_ok=True)
-
     # get tags number of occurrences
-    all_tags = {}
+    all_tags: dict[str, int] = {}
     for _, context in articles:
-        tags = context.get('tags', [])
+        tags: list[str] = context.get('tags', [])
         for tag in tags:
             all_tags[tag] = all_tags.get(tag, 0) + 1
     # sort by occurrence
-    all_tags = sorted(all_tags.items(), key=lambda x: x[1], reverse=True)
+    taglist: list[tuple[str, int]] = sorted(
+        all_tags.items(), key=lambda x: x[1], reverse=True
+    )
 
-    result = tags_template.render(dict(tags=all_tags))
+    result = tags_template.render(dict(tags=taglist))
     with open(f'{output_dir}/tags/index.html', 'w') as fh:
         fh.write(result)
 
     # get tags and archive per tag
-    all_tags = {}
+    all_tags2: dict[str, list[dict[str, Any]]] = {}
     for dst, context in articles:
         tags = context.get('tags', [])
         for tag in tags:
-            archive = all_tags.get(tag, [])
+            archive: list[dict[str, Any]] = all_tags2.get(tag, [])
             entry = context.copy()
             entry['dst'] = dst
             archive.append(entry)
-            all_tags[tag] = archive
+            all_tags2[tag] = archive
 
-    for tag, archive in all_tags.items():
+    for tag, archive in all_tags2.items():
         result = tag_template.render(dict(archive=archive, tag=tag))
         with open(f'{output_dir}/tags/{tag}.html', 'w') as fh:
             fh.write(result)
